import json
import logging
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from typing import Generator

from fastapi import FastAPI, Request
from fastapi.responses import Response
from fastapi_agmin import create_agmin
from sqlalchemy import (
    Boolean,
    CheckConstraint,
    Column,
    Date,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    LargeBinary,
    Numeric,
    String,
    Text,
    Time,
    create_engine,
)
from sqlalchemy import Enum as SAEnum
from sqlalchemy.orm import declarative_base, relationship, sessionmaker

# Configure logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

# SQLAlchemy setup
Base = declarative_base()
engine = create_engine("sqlite:///test.db")
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


# Dependency for FastAPI: yields a session and always closes it
def get_db() -> Generator:
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


# Enums
class TaskStatus(str, Enum):
    PENDING = "pending"
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    CANCELLED = "cancelled"


class PriorityLevel(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    URGENT = "urgent"


# Models
class Person(Base):
    __tablename__ = "persons"
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String(100), nullable=False)
    email = Column(String(255), unique=True, nullable=False)
    phone = Column(String(20))
    birth_date = Column(Date)
    created_at = Column(DateTime, default=datetime.utcnow)
    is_active = Column(Boolean, default=True)
    bio = Column(Text)
    profile_picture = Column(LargeBinary)
    allocations = relationship("Allocation", back_populates="person")
    address = relationship("Address", back_populates="person", uselist=False)

    def __str__(self):
        return f"{self.name} ({self.email})"


class Address(Base):
    __tablename__ = "addresses"
    id = Column(Integer, primary_key=True, index=True)
    person_id = Column(Integer, ForeignKey("persons.id"), unique=True)
    street = Column(String(200), nullable=False)
    city = Column(String(100), nullable=False)
    state = Column(String(100))
    country = Column(String(100), nullable=False)
    postal_code = Column(String(20))
    latitude = Column(Float)
    longitude = Column(Float)
    person = relationship("Person", back_populates="address")

    def __str__(self):
        return f"{self.street}, {self.city}, {self.state} {self.postal_code}"


class Task(Base):
    __tablename__ = "tasks"
    id = Column(Integer, primary_key=True, index=True)
    title = Column(String(200), nullable=False)
    description = Column(Text)
    status = Column(SAEnum(TaskStatus), default=TaskStatus.PENDING)
    priority = Column(SAEnum(PriorityLevel), default=PriorityLevel.MEDIUM)
    due_date = Column(DateTime)
    estimated_hours = Column(Float)
    actual_hours = Column(Float)
    budget = Column(Numeric(10, 2))
    created_at = Column(DateTime, default=datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)
    is_archived = Column(Boolean, default=False)
    start_time = Column(Time)
    end_time = Column(Time)
    allocations = relationship("Allocation", back_populates="task")
    attachments = relationship("Attachment", back_populates="task")

    def __str__(self):
        return f"{self.title} ({self.status.value})"


class Allocation(Base):
    __tablename__ = "allocations"
    id = Column(Integer, primary_key=True, index=True)
    person_id = Column(Integer, ForeignKey("persons.id"))
    task_id = Column(Integer, ForeignKey("tasks.id"))
    allocation_date = Column(Date)
    hours_worked = Column(Float, CheckConstraint("hours_worked IS NULL OR (hours_worked >= 2 AND hours_worked <= 60)"))
    hourly_rate = Column(Numeric(10, 2))
    notes = Column(Text)
    person = relationship("Person", back_populates="allocations")
    task = relationship("Task", back_populates="allocations")

    def __str__(self):
        person = self.person.name if self.person else "?"
        task = self.task.title if self.task else "?"
        return f"{person} - {task} ({self.allocation_date})"


class Attachment(Base):
    __tablename__ = "attachments"
    id = Column(Integer, primary_key=True, index=True)
    task_id = Column(Integer, ForeignKey("tasks.id"))
    filename = Column(String(255), nullable=False)
    file_type = Column(String(100))
    file_size = Column(Integer)  # in bytes
    upload_date = Column(DateTime, default=datetime.utcnow)
    content = Column(LargeBinary)
    task = relationship("Task", back_populates="attachments")

    def __str__(self):
        return f"{self.filename} ({self.file_type})"


def create_sample_data(session=None):
    """Create sample data for testing."""
    if session is None:
        db = SessionLocal()
    else:
        db = session

    try:
        # Create Persons
        persons = [
            Person(
                name="John Doe",
                email="john@example.com",
                phone="123-456-7890",
                birth_date=date(1990, 1, 1),
                is_active=True,
                bio="Senior Developer",
            ),
            Person(
                name="Jane Smith",
                email="jane@example.com",
                phone="098-765-4321",
                birth_date=date(1992, 5, 15),
                is_active=True,
                bio="Project Manager",
            ),
            Person(
                name="Bob Wilson",
                email="bob@example.com",
                phone="555-123-4567",
                birth_date=date(1985, 12, 31),
                is_active=False,
                bio="QA Engineer",
            ),
        ]
        db.add_all(persons)
        db.flush()  # To get IDs

        # Create Addresses
        addresses = [
            Address(
                person_id=persons[0].id,
                street="123 Main St",
                city="New York",
                state="NY",
                country="USA",
                postal_code="10001",
                latitude=40.7128,
                longitude=-74.0060,
            ),
            Address(
                person_id=persons[1].id,
                street="456 Park Ave",
                city="San Francisco",
                state="CA",
                country="USA",
                postal_code="94102",
                latitude=37.7749,
                longitude=-122.4194,
            ),
        ]
        db.add_all(addresses)

        # Create Tasks
        tasks = [
            Task(
                title="Implement User Authentication",
                description="Add OAuth2 authentication to the API",
                status=TaskStatus.IN_PROGRESS,
                priority=PriorityLevel.HIGH,
                due_date=datetime(2024, 6, 1),
                estimated_hours=40.0,
                actual_hours=25.0,
                budget=Decimal("5000.00"),
                start_time=time(9, 0),
                end_time=time(17, 0),
            ),
            Task(
                title="Database Migration",
                description="Migrate from SQLite to PostgreSQL",
                status=TaskStatus.PENDING,
                priority=PriorityLevel.MEDIUM,
                due_date=datetime(2024, 7, 1),
                estimated_hours=80.0,
                actual_hours=0.0,
                budget=Decimal("10000.00"),
            ),
            Task(
                title="Frontend Redesign",
                description="Update UI to match new brand guidelines",
                status=TaskStatus.COMPLETED,
                priority=PriorityLevel.LOW,
                due_date=datetime(2024, 5, 1),
                estimated_hours=60.0,
                actual_hours=55.0,
                budget=Decimal("7500.00"),
                is_archived=True,
            ),
        ]
        db.add_all(tasks)
        db.flush()

        # Create Allocations
        allocations = [
            Allocation(
                person_id=persons[0].id,
                task_id=tasks[0].id,
                allocation_date=date(2024, 5, 1),
                hours_worked=25.0,
                hourly_rate=Decimal("50.00"),
                notes="Working on OAuth implementation",
            ),
            Allocation(
                person_id=persons[0].id,
                task_id=tasks[1].id,
                allocation_date=date(2024, 5, 2),
                hours_worked=19.0,
                hourly_rate=Decimal("50.00"),
                notes="Working on OAuth implementation 2 (repeatead)",
            ),
            Allocation(
                person_id=persons[1].id,
                task_id=tasks[1].id,
                allocation_date=date(2024, 5, 2),
                hours_worked=None,  # Planning phase, no hours worked yet
                hourly_rate=Decimal("75.00"),
                notes="Planning phase",
            ),
            Allocation(
                person_id=persons[2].id,
                task_id=tasks[2].id,
                allocation_date=date(2024, 4, 15),
                hours_worked=55.0,
                hourly_rate=Decimal("45.00"),
                notes="Completed UI updates",
            ),
        ]
        db.add_all(allocations)

        # Create Attachments
        attachments = [
            Attachment(
                task_id=tasks[0].id,
                filename="auth_design.pdf",
                file_type="application/pdf",
                file_size=1024 * 1024,  # 1MB
                upload_date=datetime(2024, 5, 1),
            ),
            Attachment(
                task_id=tasks[1].id,
                filename="migration_plan.docx",
                file_type="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
                file_size=512 * 1024,  # 512KB
                upload_date=datetime(2024, 5, 2),
            ),
        ]
        db.add_all(attachments)

        db.commit()
    except Exception as e:
        db.rollback()
        raise e
    finally:
        db.close()


def init_db():
    """Initialize the database with tables and sample data."""
    # Drop all tables and recreate them
    Base.metadata.drop_all(bind=engine)
    Base.metadata.create_all(bind=engine)

    # Create sample data
    create_sample_data()


# Initialize database
init_db()

# FastAPI app setup
app_test = FastAPI()


@app_test.middleware("http")
async def log_requests(request: Request, call_next):
    # Log request
    body = await request.body()
    logger.debug(f"Request: {request.method} {request.url}")
    if body:
        try:
            logger.debug(f"Request body: {json.loads(body)}")
        except json.JSONDecodeError:
            logger.debug(f"Request body: {body}")

    # Get response
    response = await call_next(request)

    # Log response
    response_body = b""
    async for chunk in response.body_iterator:
        response_body += chunk

    logger.debug(f"Response status: {response.status_code}")
    if response_body:
        try:
            logger.debug(f"Response body: {json.loads(response_body)}")
        except json.JSONDecodeError:
            logger.debug(f"Response body: {response_body}")

    # Reconstruct response
    return Response(
        content=response_body,
        status_code=response.status_code,
        headers=dict(response.headers),
        media_type=response.media_type,
    )


# Initialize SQLAlchemyAgmin
create_agmin(Base, get_session=get_db, app=app_test)



import logging
import time
from datetime import date, datetime, timezone

import pytest
from fastapi.testclient import TestClient
from fastapi_agmin.agmin import StaticAssetFilter
from sqlalchemy.orm import Session

from src.fastapi_agmin.agmin import SQLAlchemyAgmin

from .app_test import Address, Allocation, Base, Person, Task


@pytest.fixture(autouse=True)
def cleanup_database(db: Session):
    """Clean up the database before each test."""
    for table in reversed(Base.metadata.sorted_tables):
        db.execute(table.delete())
    db.commit()
    yield


def test_get_metadata(client: TestClient):
    """Test getting model metadata."""
    response = client.get("/agmin/api/metadata")
    assert response.status_code == 200
    data = response.json()
    assert "models" in data
    assert "Person" in data["models"]
    assert "Task" in data["models"]
    assert "Allocation" in data["models"]
    assert "Address" in data["models"]


def test_list_persons(client: TestClient, db: Session):
    """Test listing persons."""
    # Create test data
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    persons = [
        Person(
            name="John Doe",
            email=f"john_{timestamp}@example.com",
            phone="123-456-7890",
            birth_date=date(1990, 1, 1),
            is_active=True,
            bio="Senior Developer",
        ),
        Person(
            name="Jane Smith",
            email=f"jane_{timestamp}@example.com",
            phone="098-765-4321",
            birth_date=date(1992, 5, 15),
            is_active=True,
            bio="Project Manager",
        ),
        Person(
            name="Bob Wilson",
            email=f"bob_{timestamp}@example.com",
            phone="555-555-5555",
            birth_date=date(1985, 11, 30),
            is_active=False,
            bio="QA Engineer",
        ),
    ]
    for person in persons:
        db.add(person)
    db.commit()
    db.refresh(persons[0])
    db.refresh(persons[1])
    db.refresh(persons[2])

    response = client.get("/agmin/api/person")
    assert response.status_code == 200
    data = response.json()
    assert len(data) == 3
    assert data[0]["name"] == "John Doe"
    assert data[1]["name"] == "Jane Smith"
    assert data[2]["name"] == "Bob Wilson"


def test_list_persons_performance(client, db):
    """Test listing >=1000 persons for performance and correctness."""
    # Bulk insert 1000 persons
    persons = [
        Person(
            name=f"Person {i}",
            email=f"person_{i}@example.com",
            phone="000-000-0000",
            birth_date=date(1990, 1, 1),
            is_active=True,
        )
        for i in range(1000)
    ]
    db.bulk_save_objects(persons)
    db.commit()

    start = time.time()
    response = client.get("/agmin/api/person")
    elapsed = time.time() - start

    assert response.status_code == 200
    data = response.json()
    assert len(data) == 1000
    # Check a few random records
    assert any(p["name"] == "Person 0" for p in data)
    assert any(p["name"] == "Person 999" for p in data)
    # Performance: adjust threshold as needed
    assert elapsed < 2.0, f"Listing 1000 persons took too long: {elapsed:.2f}s"


def test_create_person(client: TestClient):
    """Test creating a person."""
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    person_data = {
        "name": "John Doe",
        "email": f"john_{timestamp}@example.com",
        "phone": "123-456-7890",
        "birth_date": "1990-01-01",
        "is_active": True,
        "bio": "Senior Developer",
    }
    response = client.post("/agmin/api/person", json=person_data)
    assert response.status_code == 200
    data = response.json()
    assert data["name"] == person_data["name"]
    assert data["email"] == person_data["email"]
    assert data["phone"] == person_data["phone"]
    assert data["birth_date"] == person_data["birth_date"]
    assert data["is_active"] == person_data["is_active"]
    assert data["bio"] == person_data["bio"]


def test_create_task(client: TestClient):
    """Test creating a task."""
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    task_data = {
        "title": f"Test Task {timestamp}",
        "description": "A test task",
        "status": "pending",
        "priority": "high",
        "due_date": "2025-12-31",
    }
    response = client.post("/agmin/api/task", json=task_data)
    assert response.status_code == 200
    data = response.json()
    assert data["title"] == task_data["title"]
    assert data["description"] == task_data["description"]
    assert data["status"] == task_data["status"]
    assert data["priority"] == task_data["priority"]
    assert data["due_date"].startswith(task_data["due_date"])


def test_create_allocation(client: TestClient, db: Session):
    """Test creating an allocation."""
    # Create test data
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    person = Person(
        name="John Doe",
        email=f"john_{timestamp}@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
    )
    task = Task(
        title=f"Test Task {timestamp}",
        description="A test task",
        status="pending",
        priority="high",
        due_date=date(2025, 12, 31),
    )
    db.add(person)
    db.add(task)
    db.commit()
    db.refresh(person)
    db.refresh(task)

    allocation_data = {
        "person_id": person.id,
        "task_id": task.id,
        "allocation_date": "2025-01-01",
        "hours_worked": 8.0,
        "hourly_rate": 50.0,
        "notes": "Test allocation",
    }
    response = client.post("/agmin/api/allocation", json=allocation_data)
    assert response.status_code == 200
    data = response.json()
    assert data["person_id"] == allocation_data["person_id"]
    assert data["task_id"] == allocation_data["task_id"]
    assert data["allocation_date"] == allocation_data["allocation_date"]
    assert float(data["hours_worked"]) == allocation_data["hours_worked"]
    assert float(data["hourly_rate"]) == allocation_data["hourly_rate"]
    assert data["notes"] == allocation_data["notes"]


def test_update_person(client: TestClient, db: Session):
    """Test updating a person."""
    # Create test data
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    person = Person(
        name="John Doe",
        email=f"john_{timestamp}@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
        bio="Senior Developer",
    )
    db.add(person)
    db.commit()
    db.refresh(person)

    # Update person
    new_timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    update_data = {
        "name": "John Updated",
        "email": f"john_updated_{new_timestamp}@example.com",
        "phone": "999-999-9999",
        "bio": "Updated Bio",
    }
    response = client.put(f"/agmin/api/person/{person.id}", json=update_data)
    assert response.status_code == 200
    data = response.json()
    assert data["name"] == update_data["name"]
    assert data["email"] == update_data["email"]
    assert data["phone"] == update_data["phone"]
    assert data["bio"] == update_data["bio"]


def test_delete_person(client: TestClient, db: Session):
    """Test deleting a person."""
    # Create test data
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    person = Person(
        name="John Doe",
        email=f"john_{timestamp}@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
    )
    db.add(person)
    db.commit()
    db.refresh(person)

    # Delete person
    response = client.delete(f"/agmin/api/person/{person.id}")
    assert response.status_code == 200

    # Verify deletion
    response = client.get(f"/agmin/api/person/{person.id}")
    assert response.status_code == 404


def test_search_persons(client: TestClient, db: Session):
    """Test searching persons."""
    # Create test data
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    person = Person(
        name="John Doe",
        email=f"john_{timestamp}@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
        bio="Senior Developer",
    )
    db.add(person)
    db.commit()
    db.refresh(person)

    # Search for person
    response = client.get("/agmin/api/person/search?q=John")
    assert response.status_code == 200
    data = response.json()
    assert len(data) == 1
    assert data[0]["label"] == f"John Doe (john_{timestamp}@example.com)"


def test_relationship_loading(client: TestClient, db: Session):
    """Test loading relationships."""
    # Create test data
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    person = Person(
        name="John Doe",
        email=f"john_{timestamp}@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
    )
    task = Task(
        title=f"Test Task {timestamp}",
        description="A test task",
        status="pending",
        priority="high",
        due_date=date(2025, 12, 31),
    )
    address = Address(
        street="123 Main St",
        city="New York",
        state="NY",
        country="USA",
        postal_code="10001",
    )
    allocation = Allocation(
        person=person,
        task=task,
        allocation_date=date(2025, 1, 1),
        hours_worked=8.0,
        hourly_rate=50.0,
    )
    person.address = address
    db.add(person)
    db.add(task)
    db.add(allocation)
    db.commit()
    db.refresh(person)
    db.refresh(task)
    db.refresh(allocation)

    # Get person with relationships
    response = client.get(f"/agmin/api/person/{person.id}")
    assert response.status_code == 200
    data = response.json()
    assert data["name"] == person.name
    assert data["email"] == person.email
    assert data["address"]["street"] == address.street
    assert len(data["allocations"]) == 1
    assert data["allocations"][0]["task"]["title"] == task.title


def test_metadata_relationship_columns(client: TestClient):
    """Test that metadata includes local_columns and remote_columns for relationships."""
    response = client.get("/agmin/api/metadata")
    assert response.status_code == 200
    data = response.json()
    models = data["models"]

    # Check Allocation -> Task relationship
    allocation_rels = models["Allocation"]["relationships"]
    assert "task" in allocation_rels
    task_rel = allocation_rels["task"]
    assert "local_columns" in task_rel
    assert "remote_columns" in task_rel
    # Allocation.task_id -> Task.id
    assert task_rel["local_columns"] == ["task_id"]
    assert task_rel["remote_columns"] == ["id"]

    # Check Allocation -> Person relationship
    assert "person" in allocation_rels
    person_rel = allocation_rels["person"]
    assert person_rel["local_columns"] == ["person_id"]
    assert person_rel["remote_columns"] == ["id"]


def test_allocation_grid_columns(client: TestClient):
    """Test that the Allocation grid shows only the expected columns."""
    response = client.get("/agmin/api/metadata")
    assert response.status_code == 200
    metadata = response.json()

    allocation_columns = metadata["models"]["Allocation"]["columns"]
    allocation_relationships = metadata["models"]["Allocation"]["relationships"]

    # Check that foreign key columns are not present
    assert "person_id" not in allocation_columns
    assert "task_id" not in allocation_columns

    # Check that all expected columns are present
    expected_columns = {"display_name_", "id", "allocation_date", "hours_worked", "hourly_rate", "notes"}
    assert set(allocation_columns.keys()) == expected_columns

    # Check that relationships are present
    assert "person" in allocation_relationships
    assert "task" in allocation_relationships


def test_person_with_many_allocations(client, db):
    """Test a person with 1000+ allocations loads relationships correctly."""
    person = Person(
        name="Bulk Alloc Person",
        email="bulkalloc@example.com",
        is_active=True,
    )
    db.add(person)
    db.commit()
    db.refresh(person)

    # Create 1000 tasks and allocations
    tasks = [
        Task(
            title=f"Task {i}",
            description="Bulk task",
            status="pending",
            priority="high",
            due_date=date(2025, 12, 31),
        )
        for i in range(1000)
    ]
    db.add_all(tasks)
    db.commit()
    db.refresh(tasks[0])

    allocations = [
        Allocation(
            person_id=person.id,
            task_id=tasks[i].id,
            allocation_date=date(2025, 1, 1),
            hours_worked=8.0,
            hourly_rate=50.0,
        )
        for i in range(1000)
    ]
    db.bulk_save_objects(allocations)
    db.commit()

    response = client.get(f"/agmin/api/person/{person.id}")
    assert response.status_code == 200
    data = response.json()
    assert len(data["allocations"]) == 1000
    # Check that each allocation has a task
    for alloc in data["allocations"][:10]:  # Just sample 10 for speed
        assert "task" in alloc
        assert alloc["task"]["title"].startswith("Task")


def test_static_asset_filter():
    """Test the StaticAssetFilter to ensure it filters out static asset requests."""
    log_filter = StaticAssetFilter()

    # Create mock log records
    static_asset_record = logging.LogRecord(
        name="uvicorn.access",
        level=logging.INFO,
        pathname="",
        lineno=0,
        msg='"GET /assets/somefile.js HTTP/1.1"',
        args=(),
        exc_info=None,
    )
    non_static_asset_record = logging.LogRecord(
        name="uvicorn.access",
        level=logging.INFO,
        pathname="",
        lineno=0,
        msg='"GET /api/someendpoint HTTP/1.1"',
        args=(),
        exc_info=None,
    )

    # Assert filtering behavior
    assert not log_filter.filter(static_asset_record), "Static asset requests should be filtered out."
    assert log_filter.filter(non_static_asset_record), "Non-static asset requests should not be filtered out."


def test_serialize_instance(db: Session):
    """Test the `_serialize_instance` method for SQLAlchemy instances."""
    # Create test data
    person = Person(
        name="John Doe",
        email="john.doe@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
        bio="Senior Developer",
    )
    address = Address(
        street="123 Main St",
        city="New York",
        state="NY",
        country="USA",
        postal_code="10001",
    )
    person.address = address
    db.add(person)
    db.commit()
    db.refresh(person)

    # Initialize SQLAlchemyAgmin
    agmin = SQLAlchemyAgmin(Base, lambda: db)

    # Serialize the person instance
    serialized = agmin._serialize_instance(person)

    # Assert serialized is a dictionary
    assert isinstance(serialized, dict), "Serialized instance should be a dictionary."

    # Assert serialized data
    assert serialized["name"] == person.name
    assert serialized["email"] == person.email
    assert serialized["address"]["street"] == address.street
    assert serialized["address"]["city"] == address.city
    assert serialized["address"]["state"] == address.state
    assert serialized["address"]["country"] == address.country
    assert serialized["address"]["postal_code"] == address.postal_code


def test_generate_metadata(db: Session):
    """Test the `generate_metadata` method for SQLAlchemy models."""
    # Initialize SQLAlchemyAgmin
    agmin = SQLAlchemyAgmin(Base, lambda: db)

    # Generate metadata
    metadata = agmin.generate_metadata()

    # Assert metadata structure
    assert "Person" in metadata["models"], "Person model should be in metadata."
    assert "Task" in metadata["models"], "Task model should be in metadata."

    # Assert Person model metadata
    person_metadata = metadata["models"]["Person"]
    assert "columns" in person_metadata, "Person metadata should include columns."
    assert "relationships" in person_metadata, "Person metadata should include relationships."
    assert "name" in person_metadata["columns"], "Person columns should include 'name'."
    assert "email" in person_metadata["columns"], "Person columns should include 'email'."

    # Assert Task model metadata
    task_metadata = metadata["models"]["Task"]
    assert "columns" in task_metadata, "Task metadata should include columns."
    assert "relationships" in task_metadata, "Task metadata should include relationships."
    assert "title" in task_metadata["columns"], "Task columns should include 'title'."
    assert "description" in task_metadata["columns"], "Task columns should include 'description'."


def test_list_items(client: TestClient, db: Session):
    """Test the `list_items` endpoint for listing items and applying filters."""
    # Create test data
    person1 = Person(
        name="John Doe",
        email="john.doe@example.com",
        phone="123-456-7890",
        birth_date=date(1990, 1, 1),
        is_active=True,
    )
    person2 = Person(
        name="Jane Smith",
        email="jane.smith@example.com",
        phone="987-654-3210",
        birth_date=date(1992, 5, 15),
        is_active=False,
    )
    task1 = Task(
        title="Task 1",
        description="First task",
        status="pending",
        priority="high",
        due_date=date(2025, 12, 31),
    )
    task2 = Task(
        title="Task 2",
        description="Second task",
        status="completed",
        priority="low",
        due_date=date(2025, 11, 30),
    )
    allocation1 = Allocation(
        person=person1,
        task=task1,
        allocation_date=date(2025, 1, 1),
        hours_worked=8.0,
        hourly_rate=50.0,
    )
    allocation2 = Allocation(
        person=person2,
        task=task2,
        allocation_date=date(2025, 2, 1),
        hours_worked=6.0,
        hourly_rate=60.0,
    )
    db.add_all([person1, person2, task1, task2, allocation1, allocation2])
    db.commit()

    # Test listing persons
    response = client.get("/agmin/api/person")
    assert response.status_code == 200
    data = response.json()
    assert len(data) == 2
    assert data[0]["name"] == "John Doe"
    assert data[1]["name"] == "Jane Smith"

    # Test listing tasks
    response = client.get("/agmin/api/task")
    assert response.status_code == 200
    data = response.json()
    assert len(data) == 2
    assert data[0]["title"] == "Task 1"
    assert data[1]["title"] == "Task 2"

    # Test relationships between persons and tasks via allocations
    response = client.get(f"/agmin/api/person/{person1.id}")
    assert response.status_code == 200
    data = response.json()
    assert len(data["allocations"]) == 1
    assert data["allocations"][0]["task"]["title"] == "Task 1"

    response = client.get(f"/agmin/api/person/{person2.id}")
    assert response.status_code == 200
    data = response.json()
    assert len(data["allocations"]) == 1
    assert data["allocations"][0]["task"]["title"] == "Task 2"



import json
import logging

import pytest
from fastapi import FastAPI, Request
from fastapi.responses import Response
from fastapi.testclient import TestClient
from fastapi_agmin import create_agmin
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.pool import StaticPool

from tests.fastapi_agmin.app_test import Base, create_sample_data

# Configure logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

# SQLAlchemy setup
session_factory = None


@pytest.fixture(scope="session")
def engine():
    """Create a test database engine."""
    engine = create_engine(
        "sqlite:///:memory:",
        connect_args={"check_same_thread": False},
        poolclass=StaticPool,
        echo=True,
    )
    Base.metadata.create_all(bind=engine)
    return engine


@pytest.fixture(scope="session")
def _session_factory(engine):
    """Create a session factory."""
    global session_factory
    session_factory = scoped_session(sessionmaker(bind=engine))
    return session_factory


@pytest.fixture(scope="function")
def db(_session_factory):
    """Create a test database session."""
    session = _session_factory()
    try:
        # Clean up existing data
        for table in reversed(Base.metadata.sorted_tables):
            session.execute(table.delete())
        session.commit()

        # Create test data using the shared function
        create_sample_data(session)
        yield session
    finally:
        session.close()
        _session_factory.remove()


def get_session():
    """Get a database session for FastAPI dependency injection."""
    if session_factory is None:
        raise RuntimeError("Session factory not initialized")
    session = session_factory()
    try:
        yield session
    finally:
        session.close()


@pytest.fixture(scope="session")
def app(_session_factory):
    """Create and configure the FastAPI application."""
    app_for_test = FastAPI(title="Test App")

    @app_for_test.middleware("http")
    async def log_requests(request: Request, call_next):
        # Log request
        body = await request.body()
        logger.debug(f"Request: {request.method} {request.url}")
        if body:
            try:
                logger.debug(f"Request body: {json.loads(body)}")
            except json.JSONDecodeError:
                logger.debug(f"Request body: {body}")

        # Get response
        response = await call_next(request)

        # Log response
        response_body = b""
        async for chunk in response.body_iterator:
            response_body += chunk

        logger.debug(f"Response status: {response.status_code}")
        if response_body:
            try:
                logger.debug(f"Response body: {json.loads(response_body)}")
            except json.JSONDecodeError:
                logger.debug(f"Response body: {response_body}")

        # Reconstruct response
        return Response(
            content=response_body,
            status_code=response.status_code,
            headers=dict(response.headers),
            media_type=response.media_type,
        )

    # Initialize admin interface
    create_agmin(Base, get_session=get_session, app=app_for_test)

    return app_for_test


@pytest.fixture(scope="function")
def client(app, db):
    """Create a test client for the FastAPI application."""
    return TestClient(app)



import time
from multiprocessing import Process

import pytest
import requests
import uvicorn
from playwright.sync_api import Error, Page

from tests.fastapi_agmin.app_test import app_test


def run_server():
    uvicorn.run(app_test, port=8079)


@pytest.fixture(scope="module", autouse=True)
def server():
    proc = Process(target=run_server, args=(), daemon=True)
    proc.start()
    yield
    proc.kill()  # Cleanup after test


def get_working_url(page: Page):
    ports_to_try = [5173, 8079]
    start = time.time()
    last_error = None
    while time.time() - start < 10:
        for port in ports_to_try:
            url = f"http://localhost:{port}/agmin"
            try:
                page.goto(url, timeout=2000)
                page.wait_for_selector("select", timeout=2000)
                return url
            except Error as e:
                last_error = e
                continue
        time.sleep(0.5)
    raise RuntimeError(f"Could not connect to admin UI on ports {ports_to_try}: {last_error}")


@pytest.mark.ui
def test_model_grids_show_expected_data(page: Page):
    get_working_url(page)
    # Use url for navigation if needed
    # page.goto(url)  # Already navigated in get_working_url

    # Model: Task
    page.select_option("select", label="Task")
    page.wait_for_timeout(500)
    headers = page.locator(".ag-header-cell-text").all_text_contents()
    assert "title" in [h.lower() for h in headers]
    assert "status" in [h.lower() for h in headers]
    first_row = page.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    print("Task first_row:", first_row)
    assert any("Implement User Authentication" in cell for cell in first_row)
    assert any("in_progress" in cell for cell in first_row)

    # Model: Person
    page.select_option("select", label="Person")
    page.wait_for_timeout(500)
    headers = page.locator(".ag-header-cell-text").all_text_contents()
    assert "name" in [h.lower() for h in headers]
    assert "email" in [h.lower() for h in headers]
    first_row = page.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    print("Person first_row:", first_row)
    assert any("John Doe" in cell for cell in first_row)
    assert any("john@example.com" in cell for cell in first_row)

    # Model: Address
    page.select_option("select", label="Address")
    page.wait_for_timeout(500)
    headers = page.locator(".ag-header-cell-text").all_text_contents()
    assert "street" in [h.lower() for h in headers]
    first_row = page.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    assert any("123 Main St" in cell for cell in first_row)

    # Model: Allocation
    page.select_option("select", label="Allocation")
    page.wait_for_timeout(500)
    headers = page.locator(".ag-header-cell-text").all_text_contents()
    assert "hours_worked" in [h.lower() for h in headers]
    first_row = page.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    assert len(first_row) > 0

    # Model: Attachment
    page.select_option("select", label="Attachment")
    page.wait_for_timeout(500)
    headers = page.locator(".ag-header-cell-text").all_text_contents()
    assert "filename" in [h.lower() for h in headers]
    first_row = page.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    assert any("auth_design.pdf" in cell for cell in first_row)


@pytest.mark.ui
def test_first_column_is_display_name_and_not_empty(page: Page):
    get_working_url(page)
    # Get all model options except the placeholder
    model_options = page.locator("select option").all()
    for option in model_options:
        model = option.text_content()
        if not model or model.strip().lower() == "select a model" or set(model) == {"─"}:
            print(f"Skipping model option: {model!r}")
            continue
        page.select_option("select", label=model)
        page.wait_for_timeout(2000)
        headers = page.locator(".ag-header-cell-text").all_text_contents()
        print(f"Model: {model}, Headers: {headers}")
        if not headers or not headers[0]:
            print("Page content:", page.content())
            continue
        assert (
            headers[0] and "display" in headers[0].lower()
        ), f"First column for {model} is not display name: {headers}"
        # Assert the second column is not Display Name
        if len(headers) > 1:
            assert headers[1] not in [
                "display_name_",
                "Display Name",
            ], f"Second column for model {model} should not be display_name_ or Display Name: {headers[1]}"
        # Get the first row's first cell value
        first_cell = (
            page.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").first.text_content()
        )
        assert first_cell and first_cell.strip(), f"First cell in display_name_ column for model {model} is empty"


@pytest.mark.ui
def test_person_detail_shows_related_entities(page: Page):
    url = get_working_url(page)
    # Get first Person's id from API
    api_url = url.replace("/agmin", "/agmin/api/person")
    resp = requests.get(api_url)
    resp.raise_for_status()
    people = resp.json()
    assert people, "No Person found in API!"
    # Go to agmin first to load the app
    page.goto(f"{url}")
    page.wait_for_timeout(500)
    page.select_option("select", label="Person")
    page.wait_for_timeout(2000)
    # Go to Person detail view
    page.locator(".relations-button").first.click()
    page.wait_for_timeout(2000)
    # Verify main Person table headers
    headers = page.locator(".ag-header-cell-text").all_text_contents()
    assert "name" in [h.lower() for h in headers]
    assert "email" in [h.lower() for h in headers]
    # Get all tables on the page
    tables = page.locator(".model-grid").all()
    assert len(tables) == 3, "Expected 3 tables: Person, Allocation, and Address"
    # Verify Person table (first table)
    person_headers = tables[0].locator(".ag-header-cell-text").all_text_contents()
    assert "name" in [h.lower() for h in person_headers]
    assert "email" in [h.lower() for h in person_headers]
    person_row = (
        tables[0].locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    )
    assert any("John Doe" in cell for cell in person_row)
    # Find and verify Allocation table
    allocation_table = None
    for table in tables[1:]:
        headers = table.locator(".ag-header-cell-text").all_text_contents()
        if "hours_worked" in [h.lower() for h in headers]:
            allocation_table = table
            break
    assert allocation_table, "Allocation table not found"
    allocation_row = (
        allocation_table.locator(".ag-center-cols-container .ag-row")
        .first.locator(".ag-cell-value")
        .all_text_contents()
    )
    assert len(allocation_row) > 0

    # Find and verify Address table
    address_table = None
    for table in tables[1:]:
        headers = table.locator(".ag-header-cell-text").all_text_contents()
        if "street" in [h.lower() for h in headers]:
            address_table = table
            break
    assert address_table, "Address table not found"
    address_row = (
        address_table.locator(".ag-center-cols-container .ag-row").first.locator(".ag-cell-value").all_text_contents()
    )
    assert any("123 Main St" in cell for cell in address_row)



import logging
from datetime import date, datetime, time
from decimal import Decimal
from functools import lru_cache
from importlib import resources
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type, cast

import yaml
from fastapi import APIRouter, Depends, FastAPI, HTTPException, Query, Request
from fastapi.responses import FileResponse, JSONResponse, RedirectResponse, Response
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel, ConfigDict, create_model
from sqlalchemy import Date, String, or_, select
from sqlalchemy import Enum as SAEnum
from sqlalchemy import inspect as sa_inspect
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import DeclarativeMeta, Mapper, Session, selectinload

DISPLAY_NAME_FIELD = "display_name_"

MAX_RECORDS = 1000


def create_agmin(
    base: DeclarativeMeta,
    get_session: Callable[[], Session],
    app: Optional[FastAPI] = None,
) -> "SQLAlchemyAgmin":
    """Create and configure a SQLAlchemyAgmin instance.

    Args:
        base: SQLAlchemy declarative base containing model definitions
        get_session: Callable that returns a SQLAlchemy session
        app: Optional FastAPI application instance

    Returns:
        SQLAlchemyAgmin: Configured admin interface instance
    """
    return SQLAlchemyAgmin(base, get_session, app)


# Configure logging filter
class StaticAssetFilter(logging.Filter):
    def filter(self, record):
        # Filter out static asset requests from uvicorn access logs
        if record.getMessage().startswith('"GET /assets/') or record.getMessage().startswith('"GET /static/'):
            return False

        # Filter out response body logs for static assets
        if "Response body:" in record.getMessage() and (
            "/assets/" in record.getMessage() or "/static/" in record.getMessage()
        ):
            return False

        return True


# Apply filter to both uvicorn access logger and test_app logger
logging.getLogger("uvicorn.access").addFilter(StaticAssetFilter())
logging.getLogger("test_app").addFilter(StaticAssetFilter())


class SQLAlchemyAgmin:
    def __init__(self, base: DeclarativeMeta, get_session, app: Optional[FastAPI] = None):
        # base is sqlalchemy.orm.declarative_base
        self.base = base
        self.get_session = get_session  # FastAPI dependency for DB session
        self.app = app

        # Build a name → model class map using SQLAlchemy 2.x compatible approach
        self.model_map = {mapper.class_.__name__.lower(): mapper.class_ for mapper in self.base.registry.mappers}

        # If app is provided, set up routes and static files
        if app is not None:
            self._setup_app(app)

    def _get_static_dir(self) -> Path:
        """Get the path to the static directory in the installed package."""
        return resources.files("fastapi_agmin").joinpath("static")

    def _setup_app(self, app: FastAPI):
        """Set up routes and static files for the FastAPI app."""
        # Include the API router
        app.include_router(self.create_fastapi_router(), prefix="/agmin/api")

        # Get the static directory path
        static_dir = self._get_static_dir()

        # Serve the Vue app's assets directory
        app.mount(
            "/assets",
            StaticFiles(directory=str(static_dir / "assets")),
            name="assets",
        )

        # Handle all /agmin/* routes
        @app.get("/agmin/{path:path}")
        async def admin_page(path: str = ""):
            if path.startswith("assets/"):
                # Let the static files handler take care of assets
                return RedirectResponse(url=f"/assets/{path[7:]}")

            # For any other path, serve index.html with next parameter
            if path:
                return RedirectResponse(url=f"/agmin/?next=/agmin/{path}")

            # For /agmin/, serve index.html directly
            return FileResponse(str(static_dir / "index.html"))

    def _serialize_instance(self, instance: Any, depth: int = 0) -> Dict[str, Any] | None:
        """Serialize a SQLAlchemy instance to a dictionary."""
        if instance is None:
            return None

        # Prevent infinite recursion
        if depth > 2:
            return {"id": getattr(instance, "id", None), DISPLAY_NAME_FIELD: str(instance)}

        data = {}
        mapper = sa_inspect(instance.__class__)

        # Add DISPLAY_NAME_FIELD field
        data[DISPLAY_NAME_FIELD] = str(instance)

        # Add columns
        for column in mapper.columns:
            value = getattr(instance, column.key)
            if isinstance(value, (date, datetime, time)):
                value = value.isoformat()
            elif isinstance(value, Decimal):
                value = str(value)
            data[column.key] = value

        # Add relationships
        for rel in mapper.relationships:
            value = getattr(instance, rel.key)
            if value is not None:
                if rel.uselist:
                    data[rel.key] = [self._serialize_instance(item, depth + 1) for item in value]
                else:
                    data[rel.key] = self._serialize_instance(value, depth + 1)
            else:
                data[rel.key] = None

        return data

    @lru_cache(maxsize=1)  # noqa: B019
    def generate_metadata(self) -> Dict[str, Any]:
        metadata = {}
        for cls in self.base.registry.mappers:
            if isinstance(cls.class_, type) and hasattr(cls.class_, "__tablename__"):
                model_info = {
                    "table": cls.class_.__tablename__,
                    "columns": {},
                    "relationships": {},
                }
                mapper = sa_inspect(cls.class_)
                # Add synthetic display_name_ column
                model_info["columns"][DISPLAY_NAME_FIELD] = {
                    "type": "VARCHAR",
                    "nullable": False,
                    "primary_key": False,
                    "unique": False,
                    "enum_values": None,
                    "label": "Display Name",
                    "help_text": "Human-readable display name for this row",
                }

                # Get foreign key columns from relationships
                fk_columns = set()
                for rel in mapper.relationships:
                    for col in rel.local_columns:
                        fk_columns.add(col.name)

                # Columns
                for column in mapper.columns:
                    # Skip columns that are foreign keys in relationships
                    if column.name in fk_columns:
                        continue

                    col_type = str(column.type)
                    enum_values = None
                    if isinstance(column.type, SAEnum):
                        enum_values = list(column.type.enums)

                    model_info["columns"][column.name] = {
                        "type": col_type,
                        "nullable": column.nullable,
                        "primary_key": column.primary_key,
                        "unique": column.unique,
                        "enum_values": enum_values,
                        "label": getattr(column, "info", {}).get("label"),
                        "help_text": getattr(column, "info", {}).get("help_text"),
                    }

                # Relationships
                for rel in mapper.relationships:
                    rel_type = (
                        "many_to_one"
                        if rel.direction.name == "MANYTOONE"
                        else (
                            "one_to_many"
                            if rel.direction.name == "ONETOMANY"
                            else ("many_to_many" if rel.direction.name == "MANYTOMANY" else "unknown")
                        )
                    )
                    local_cols = [col.name for col in rel.local_columns]
                    remote_cols = [col.name for col in rel.remote_side]
                    relationship_info = {
                        "type": rel_type,
                        "target": rel.mapper.class_.__name__,
                        "back_populates": rel.back_populates,
                        "secondary": (rel.secondary.name if rel.secondary is not None else None),
                        "local_columns": local_cols,
                        "remote_columns": remote_cols,
                    }
                    model_info["relationships"][rel.key] = relationship_info

                metadata[cls.class_.__name__] = model_info

        return {"models": metadata}

    def _create_pydantic_model(
        self,
        cls: Type,
        include_relationships: bool = False,
        exclude_readonly: bool = False,
    ) -> Type[BaseModel]:
        """
        Dynamically create a Pydantic model from a SQLAlchemy model.

        :param cls: SQLAlchemy model class
        :param include_relationships: Whether to include relationships in the model
        :param exclude_readonly: Whether to exclude read-only fields (e.g., primary keys)
        :return: Pydantic model class
        """
        fields = {}
        mapper = sa_inspect(cls)

        # Add DISPLAY_NAME_FIELD field for display
        fields[DISPLAY_NAME_FIELD] = (str, "")

        # Add columns
        for column in mapper.columns:
            if exclude_readonly and column.primary_key:
                continue
            python_type = column.type.python_type
            if column.nullable:
                python_type = Optional[python_type]
            fields[column.key] = (python_type, None if column.nullable else ...)

        # Add relationships if required
        if include_relationships:
            for rel in mapper.relationships:
                if rel.uselist:
                    fields[rel.key] = (List[Any], [])
                else:
                    fields[rel.key] = (Optional[Dict[str, Any]], None)

        return create_model(
            f"{cls.__name__}Model",
            **fields,
            __config__=ConfigDict(from_attributes=True),
        )

    def create_fastapi_router(self) -> APIRouter:
        router = APIRouter()
        db_dependency = Depends(self.get_session)

        @router.get("/metadata", response_model=Dict[str, Any])
        async def get_metadata(format: str = Query("json", enum=["json", "yaml"])):  # noqa: A002
            metadata = self.generate_metadata()
            if format == "yaml":
                yaml_str = yaml.dump(metadata, sort_keys=False, allow_unicode=True)
                return Response(content=yaml_str, media_type="application/x-yaml")
            else:
                return JSONResponse(content=metadata)

        @router.get("/{model}/search")
        async def search_items(
            model: str,
            q: str = Query(..., min_length=3),
            db: Session = db_dependency,
        ) -> List[Dict[str, Any]]:
            cls = self._get_model(model)
            mapper = sa_inspect(cls)

            # Build search conditions for string columns
            conditions = []
            for column in mapper.columns:
                if isinstance(column.type, (String, Date)):
                    conditions.append(column.ilike(f"%{q}%"))

            if not conditions:
                return []

            query = select(cls).where(or_(*conditions)).limit(MAX_RECORDS)
            items = db.scalars(query).all()

            # Format results using __str__ method
            return [{"id": item.id, "label": str(item)} for item in items]

        @router.get("/{model}")
        async def list_items(
            model: str,
            db: Session = db_dependency,
            request: Request = None,
        ) -> List[Dict[str, Any]]:
            filters = dict(request.query_params)
            cls = self._get_model(model)
            output_model = self._create_pydantic_model(cls, include_relationships=True)
            query = select(cls)
            mapper = sa_inspect(cls)
            # Foreign key filtering
            for rel in mapper.relationships:
                rel_param = f"{rel.key}_id"
                if rel_param in filters and filters[rel_param] is not None:
                    try:
                        fk_id = int(filters[rel_param])
                    except Exception:
                        continue
                    # For many-to-one, filter by foreign key column
                    if rel.direction.name == "MANYTOONE":
                        fk_col = list(rel.local_columns)[0]
                        query = query.where(fk_col == fk_id)
                    # For one-to-many, filter by existence of related row
                    elif rel.direction.name == "ONETOMANY":
                        related_cls = rel.mapper.class_
                        # Use rel.primaryjoin to get the join condition between parent and related
                        join_cond = rel.primaryjoin
                        subq = select(1).where((related_cls.id == fk_id) & join_cond).exists()
                        query = query.where(subq)
                    # For many-to-many, skip for now
            for rel in mapper.relationships:
                query = query.options(selectinload(getattr(cls, rel.key)))
            query = query.limit(MAX_RECORDS)
            items = db.scalars(query).all()
            result = [output_model.model_validate(self._serialize_instance(item)).model_dump() for item in items]
            return result

        @router.get("/{model}/{item_id}")
        async def get_item(
            model: str,
            item_id: int,
            db: Session = db_dependency,
        ) -> Dict[str, Any]:
            cls = self._get_model(model)
            output_model = self._create_pydantic_model(cls, include_relationships=True)

            # Build query with eager loading for all relationships
            query = select(cls).where(cls.id == item_id)
            for rel in sa_inspect(cls).relationships:
                # Load all relationships eagerly
                query = query.options(selectinload(getattr(cls, rel.key)))

            item = db.scalars(query).first()
            if not item:
                raise HTTPException(status_code=404, detail="Item not found")
            result = output_model.model_validate(self._serialize_instance(item)).model_dump()
            result[DISPLAY_NAME_FIELD] = str(item)
            return result

        @router.post("/{model}")
        async def create_item(
            model: str,
            payload: Dict[str, Any],
            db: Session = db_dependency,
        ) -> Dict[str, Any]:
            cls = self._get_model(model)
            input_model = self._create_pydantic_model(cls, exclude_readonly=True)
            output_model = self._create_pydantic_model(cls, include_relationships=True)
            validated_data = input_model(**payload)
            item = cls()

            # Handle relationships first
            mapper = sa_inspect(cls)
            for rel in mapper.relationships:
                if rel.key in validated_data.model_dump():
                    value = validated_data.model_dump()[rel.key]
                    if value is not None:
                        related_cls = rel.mapper.class_
                        if rel.uselist:
                            related_objs = db.query(related_cls).filter(related_cls.id.in_(value)).all()
                            setattr(item, rel.key, related_objs)
                        else:
                            related_obj = db.get(cast(Mapper, related_cls), value)
                            setattr(item, rel.key, related_obj)

            # Handle regular fields
            for key, value in validated_data.model_dump().items():
                if key not in [rel.key for rel in mapper.relationships]:
                    setattr(item, key, value)

            db.add(item)
            db.commit()
            db.refresh(item)
            return output_model.model_validate(self._serialize_instance(item)).model_dump()

        @router.put("/{model}/{item_id}")
        async def update_item(
            model: str,
            item_id: int,
            payload: Dict[str, Any],
            db: Session = db_dependency,
        ) -> Dict[str, Any]:
            cls = self._get_model(model)
            input_model = self._create_pydantic_model(cls, exclude_readonly=True)
            output_model = self._create_pydantic_model(cls, include_relationships=True)
            validated_data = input_model(**payload)
            item = db.get(cast(Mapper, cls), item_id)
            if not item:
                raise HTTPException(status_code=404, detail="Item not found")

            try:
                # Handle relationships first
                mapper = sa_inspect(cls)
                for rel in mapper.relationships:
                    if rel.key in validated_data.model_dump():
                        value = validated_data.model_dump()[rel.key]
                        if value is not None:
                            related_cls = rel.mapper.class_
                            if rel.uselist:
                                related_objs = db.query(related_cls).filter(related_cls.id.in_(value)).all()
                                setattr(item, rel.key, related_objs)
                            else:
                                related_obj = db.get(cast(Mapper, related_cls), value)
                                setattr(item, rel.key, related_obj)

                # Handle regular fields
                for key, value in validated_data.model_dump().items():
                    if key not in [rel.key for rel in mapper.relationships]:
                        setattr(item, key, value)

                db.commit()
                db.refresh(item)
                return output_model.model_validate(self._serialize_instance(item)).model_dump()
            except IntegrityError as e:
                db.rollback()
                error_msg = str(e.orig)
                if "CHECK constraint failed" in error_msg:
                    # Extract the field name from the error message
                    field_name = error_msg.split(":")[-1].strip()
                    raise HTTPException(
                        status_code=422,
                        detail={
                            "loc": ["body", field_name],
                            "msg": f"Invalid value: {error_msg}",
                            "type": "value_error",
                        },
                    ) from e
                raise HTTPException(status_code=400, detail=str(e)) from e

        @router.delete("/{model}/{item_id}", response_model=Dict[str, str])
        async def delete_item(
            model: str,
            item_id: int,
            db: Session = db_dependency,
        ):
            cls = self._get_model(model)
            item = db.get(cast(Mapper, cls), item_id)
            if not item:
                raise HTTPException(status_code=404, detail="Item not found")
            db.delete(item)
            db.commit()
            return {"detail": "Item deleted"}

        return router

    def _get_model(self, model_name: str):
        cls = self.model_map.get(model_name.lower())
        if not cls:
            raise HTTPException(status_code=404, detail=f"Model '{model_name}' not found")
        return cls
